/*
    MSGLIB - a message handling library
    Copyright (C) 1995,1997  Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    See: COPYING.LB
*/
/*
    MSGCOMP2 - utility for the message handling library
    Copyright (C) 1995,97  Steffen Kaiser

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
*/
/* $RCSfile: MSGCOMP2.C $
   $Locker:  $	$Name:  $	$State: Exp $

	The message compiler 2nd generation.

   ----
	command line usage:

	msgComp2 [{option}] pathGlbMsg lng locHeader [msgDcl [msgFeatures]]


	pathGlbMsg  : path to the global message definition files
	lng			: language to be compiled (message description file)
	locHeader	: local header file that includes the local msg decl file
	msgDcl		: message declarations
	msgFeature	: list of features used within the description file

	If the extension of <lng> is omited, ".MSG" is assumed.

	All local messages are read from <lng>.
	Internally used names are read from both:
			<pathGlbMsg>\<lng>[.msg]
	and		<pathGlbMsg>\english.msg
	But if one word resides in both files, the meaning out of
	<lng>[.msg] takes precedance.
	In the english.msg message description file the internally used
	names are of the form:
		>>>>>>I_internalKeywords
		<number of keywords>
		<keyword> | <comment>
	The <number of keywords> defines how many keywords exist.
	A <comment> is any line that does not start with a digit.
	A <keyword> definition looks like this:
		<numerical ID> <symbolical ID> <string> \n
	The <numerical ID> is the number that is internally used to identify
	this keyword. It must be equal to the #define'ed macro 'IN_????'.
	The <symbolic ID> is the word with what the English keywords are
	linked to the language specific keywords. The <symbolic ID> must
	start with 'IN_'!
	The rest of the line is interpreted as the <string> that identifies
	the keyword. It may contain spaces.
	The I_internalKeywords message may appear more than once within the
	file, but the first line will be ignored except for the first one.

	The message compiler creates numerous message definition files
	"msg_####.c" containing the message string definitions and the
	header file (msgFile with the extension ".msg") containing the
	message string declarations, the error level definitions, and
	the mapping to the native language (see below).
	These files are garantied to be plain 8-bit ASCII. All characters
	00 through 0x1F are represented by their octal form, e.g.:
		ASCII(24) -> \30
	While all characters except NUL (ASCII(0)) are translated immediately
	before writing them into the output file, NUL is translated while
	reading. This is necessary, because NUL is internally used as the string
	terminator (well, that's the C style); there is only one problem I can
	see so far: The %--%A and %--%W specials treats the next character as
	the delimiter; if one choose the NUL here ...

	Options:
		/!	switches the name to "msg!####", because VAL complains about
			multiple modules. So the global messages can be generated
			with this switch.
			This also implies /g=8.
		/b	open the message description files in binary mode. That allows
			to include '\0's and ^Z characters in the file. Although there
			will be made no translation of the "newline" character!
		/g=	Assigns a message group prefix. Range 0..15. This is the
			most-highest half-byte of the message number.
			Currently assigned groups:
				0: local message
				8: global message
				others: reserved for future use
		/l=	causes to output the library response file into the specified
			file. This file looks like:
				+- msg_0000.obj &
				+- msg_0001.obj &
				   ...
				+- msg_####.obj
		/c=	causes to output the compiler response file into the specified
			file. This file looks like:
				msg_0000.c
				msg_0001.c
				   ...
				msg_####.c
		/m=	causes to output a make dependiency structure into the
			specified file. This file looks like:
				msg_0000.obj : msg_0000.c
				msg_0001.obj : msg_0001.c
				   ...
				msg_####.obj : msg_####.c


   Files:
   	All files will be overwritten without further notice. If they are
   	preceeded by a '+' sign, the new information will be appended.

   ----

   Target compilers:
   		Micro-C v3.14, Borland C++ v3.1+
   ----

	The message declarations:

	To declare a message a macro is provided. That macro has three
	parameters:
		DCL_MSG(msgId, msgGroup, msgNr)

	The msgId is the source code identification of the message as it
	apears in the message head of a message definition file.
	The msgGroup is the group, in which the message is located (see /g
	option).
	The msgNr is a number, that uniquely identifies the message ID within
	its message group. The current implementation allows up to 4096
	(0xFFF) message per group.

   ----

   	The message description file syntax:

   	Each message consists of a message head and a message body. The message
   	head is the whole line of the format: ">>>>>>?_name [#]\n". The message
   	body consists of all lines up to the next message header or the EOF.

   	name :: is the name of the message in English
   	?    :: is the message group and must be a single upper-cased letter
   	#    :: error level (for ? == 'E' only)

	The message body may contain special strings: "%-%*%", where '*' stands
	for a comma delimited list of the following ANSI keywords:
		cls			clear screen
		clreol		clear from cursor to end of line
		
		home		cursor to upper left corner
		up #		cursor up # rows		\
		down #		cursor down # rows		 | neither scroll nor change
		left #		cursor left # columns	 | of the other coordiante
		right #		cursor right # columns	/
		goto # #	move cursor to position row,column (one based!!)

		save		save one cursor position
		restore		restore saved cursor position

	  colour related keywords:
		bright		set attribute: foreground colour highlighted
		blink		set attribute: blinking character
		inverse		set attribute: inverse
		underline	set attribute: underline
		hidden		set attribute: hidden
		normal		unset all attributes
		black		foreground colour: black 
		red			foreground colour: red 
		green		foreground colour: green 
		yellow		foreground colour: yellow 
		blue		foreground colour: blue 
		magenta		foreground colour: magenta 
		cyan		foreground colour: cyan 
		white		foreground colour: white 

		Those keywords may be combined in any order WITHOUT a delimiting
		comma. If more than one colour is given, the first is treated as
		a foreground colour, all others as background colours.

	These keywords must be defined in the global message file within
	message group 'I' == internal or in the global English message file.


	The special string: %--% includes the following functionality:
		%--%A?****?
			ANSI enclosure. **** is any string that is emitted, if the
			output of ANSI sequences is allowed. ? is any character that
			is not contained within ****. The string may be longer than
			255 characters, but it will be broken into pieces smaller
			than 256 characters.
		%--%W?****?
			Raw data enclosure; as %--%A, but the string is always emitted.
		%--%r?###%
			Output the single character ? ### times. Although ### may be
			greater than 255, it will be broken into pieces less than
			256.
		%--%R###%
			As above, but the character to be emitted is an argument of the
			type (int).
		%--%m****%
			Recursively include the message with the ID ****. The text of
			the message **** is interpreted, as if it would replace this
			special token, that means the "%1%fmt" means the same argument
			in both the including and the included message.
		%--%M
			As above, but the message to be included is specified by one
			argument of the type (MSGID). Because the included message may
			also contain argument specifications all these argument should
			preceed all other arguments in both conjunctions the argument
			type declaration and as parameter, e.g.:
				>>>>>>M_include
				%-1%MAny arguments %s %d, and the message %1%M, and other
				arguments %s %c.
				>>>>>>M_includedMessage
				%s

				To output the message use that parameter sequence:
					<message to be included>, <para of 1st %s>, <para of %d>,
					<para of %s of M_includedMessage>,
					<para of 2nd %s>, <para of %c>
				e.g.:
					smessage(M_include, M_includedMessage, "s1", 1,
						"s of included msg", "s2", 'c');
				If M_includedMessage would contain "%1%s", "s1" would
				be emitted.

   ----

	Because the message body is transformed into a C string definition,
	any non-printable characters are replaced by their C-style and already
	enterred C-style is interpreted by the C compiler, the last '\n' is
	stripped, e.g.:

>>>>>>M_dummy
	line\\	1
	line\r	2
>>>>>>

	is transformed into:

	"\tline\\\t1\n\tline\r\t2"


   ----
	
	Known bugs:
		Most compilers have problems with '\0' characters within
		the IN stream. Therefore, it's not generally possible to
		place them e.g. in the %#-%??? special. Also, because of
		the text mode reading '\x1a' (^Z) is not allowed as a
		legal character.

		The string attached to the %#-%??? special is not
		processed in any way, especially no C-style syntax is
		translated here.

   ----
	
*/

#include <stdio.h>
#ifndef _MICROC_
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#endif
#include <portable.h>
#include <dfn.h>
#include <getopt.h>

#ifdef __TURBOC__
#pragma hdrstop
#endif

#include "yerror.h"

#ifndef NUL
#define NUL '\0'
#endif
#define BUF 2048 	/* buffer size while copying a file */

#include "keywords.inc"

#define MSG_ID_INTERNAL 'I'		/* prefix of internal used messages */
#define MSG_ID_ERROR 'E'		/* prefix of error messages */
#define NO_NR 0xFFFF
#define GLOBAL_GROUP 8		/* group of global messages */

#ifndef _MICROC_
#define appendMode "at"
#else
#define appendMode "wat"
#endif

#ifndef lint
static char const rcsid[] = 
	"$Id: MSGCOMP2.C 1.8 1998/01/14 08:31:59 ska Exp $";
#endif

char fmode[] = "rt";						/* open mode for message file */


/* opening of all message definition files */
const char msgOpening[] = "\
#ifndef _MICROC_\n\
#include <algnbyte.h>\n\
#else \n\
#define const \n\
#endif\n\
#include \"%s\"\n";						/* local message header file */
										/* needed for message includes */

const char msgErrOpening[] = "\
static unsigned char errNr = %u;\n";	/* message error number */

const char msgMsgDef[] = "\
const unsigned char %s[] = \"\\\n";		/* message head */

/* closing of all message definition files */
const char msgClosing[] = "\
\";\n";

/* Id of special enclosures */
const char specId[] = "\
%%-%%%c";

/* include of a message id:
	1. line: finish previous string portion
	2. line: include message ID
	3. line: declare next string portion
	The 2. and 3. lines are unified by contiguous numbers to allow
	more than one message include. */
const char msgInclude[] = "\
\";\\\n\
#include \"%s\"\\\n\
const static MSGID mid%u = %s;\\\n\
const static unsigned char sid%u[] = \"\\\n";

/* how to declare a message */
char msgDCL[] = "DCL_MSG(%s, %u, %u)\n";

/*
 *	Buffer for internal commands
 *	internal[] is for language specific keywords
 *	internalE[] is for English keywords
 *
 *	The keywords are identified in the English message description file
 *	then read from the language specific description file.
 */
char **internal = NULL;
char **internalE = NULL;
unsigned keywords = 0;
const char iKeywords[] = "I_internalKeywords";
const char symID[] = "IN_";			/* how symbolic IDs start */


/*
 *	default extension for the message description files
 */
const char extMsg[] = "MSG";
const char extDcl[] = "DCL";
const char extFtr[] = "FTR";
const char fnMsg[] = "ENGLISH";


char msgDefFile[] = "%smsg_%04x%s";	/* name of the message definition files */
#define padDefFile 5		/* offset of '_' */

char *msgFile; /* extension-less local message file; with space for ext */
char *msgBody = NULL;			/* current message body */
unsigned msgIncCnt = 0;			/* counter within 1 msg def file */
char *locHead;					/* local header file */
unsigned msgDefFileNr = 0;	/* number of message definition files */
char msgDefFileStr[13];

unsigned errNr = 0;			/* error level of current message header */
unsigned lineNr = 0;		/* keep it NOT (long) because of Micro-C */
FLAG binary = 0;
int dclCnt = 0;				/* message counter */
int group = 0;				/* message group */

/*	Flags, which features has been used */
const char iFeatures[] = "I_features";

#define featAnsi features[0]					/* %-%A */
#define featRaw features[1]						/* %-%W */
#define featRep features[2]						/* %-%r */
#define featRepArg features[3]					/* %-%R */
#define featMInc features[4]					/* %-%m */
#define featMIncArg features[5]					/* %-%M */
#define featReorder features[6]					/* %-#%fmt or %#%fmt */
#define featNUL features[7]						/* '\0' within message */
#define FEATURES 8

FLAG features[FEATURES] = {0};
char *featDef[FEATURES] = {
	"#define MSG_FEAT_ANSI",
	"#define MSG_FEAT_RAW",
	"#define MSG_FEAT_REP",
	"#define MSG_FEAT_REPA",
	"#define MSG_FEAT_MINC",
	"#define MSG_FEAT_MINCA",
	"#define MSG_FEAT_REORDER",
	"#define MSG_FEAT_NUL" };



/*
 *	Allocate a chunk of memory
 *	On failure: Terminate with an error message
 */
byte *getmem(int len)
{	byte *h;

	if((h = malloc(len)) == NULL)
		error(E_noMem);

	return h;
}


/*
 *	Allocate a chunk of memory and erase the contents to zero
 *	On failure: Termiante with an error message
 */
byte *cgetmem(int len)
{	byte *h;

	memset(h = getmem(len), 0, len);
	return h;
}



/*
 *	Duplicate a string into the local heap.
 *	On failure: Terminate with an error message
 */
char *dupstr(char *str)
{	char *h;

	if((h = strdup(str)) == NULL)
		error(E_noMem);

	return h;
}


/*
 *	Duplicate a portion of a string
 */
char *strdupe(char *start, char *end)
{	int c;

	c = *end;
	*end = NUL;
	start = dupstr(start);
	*end = c;

	return start;
}


/*
 *	StrCat() within dynamic memory
 */
#define addStr(a,b) addStr_(&(a),(b))
char *addStr_(char **str, char *s)
{
	if(s == NULL) return *str;
	if(*str == NULL)
		return *str = dupstr(s);

	if((*str = realloc(*str, strlen(*str) + strlen(s) + 1)) == NULL)
		fatal(E_noMem);
#ifdef _MICROC_
	stpcat(*str, s);
	return *str;
#else
	return strcat(*str, s);
#endif
}


/*
 *	Skip whitespaces
 */
char *skipws(char *s)
{	while(isspace(*s)) ++s;
	return s;
}


/*
 *	Skip to the end of word (one byte behind the word)
 */
char *eow(char *s)
{	while(*s && *s != '\n' && !isspace(*s)) ++s;
	return s;
}

/*
 *	Skip to the end of line (trailing spaces ignored)
 */
char *eol(char *s)
{	char *e;

	if((e = strchr(s, '\n')) == NULL)
		e = strchr(s, NUL);
	while(--e >= s && isspace(*e));
	return e + 1;
}


/*
 *	Open a file
 *	On failure terminate with an error message
 */
FILE *Efopen(char *fname, const char *mode)
{	FILE *fp;

	/* open file & emit error on failure */
	if((fp = fopen(fname, mode)) == NULL)
		fatal(E_openFile, fname);

	return fp;
}

/*
 *	Open a file with mode "wt", unless its first character is '+'.
 *	In that case open the file with mode "at".
 */
FILE *Efopen1(char *fn)
{	iM(FILE *f;)

	if(*fn == '+') {
#ifndef _MICROC_
		return Efopen(fn + 1, appendMode);
#else
	/* Micro-C has a problem, when the file does not already exist */
	if(f = fopen(fn + 1, appendMode))
		return f;

	++fn;			/* if failed -> use "wt" mode */
#endif
	}

	return Efopen(fn, "wt");
}


/*
 *	Close a file
 *	If an error occured while writing/reading the file, terminate
 *	with an error message.
 */
void Efclose(FILE *fp, char *fnam)
{	/* ensure, fp->buffer is empty, so ferror() is up to date */
	fflush(fp);

	if(ferror(fp))
		fatal(E_writeFile, fnam);

	fclose(fp);
}


/*
 *	Makes sure that the file has an extension attached.
 */
char *attachExt(char *fn, const char * const ext)
{	char *dr, *pa, *fi, *ex;

	if(!dfnsplit(fn, &dr, &pa, &fi, &ex)
	 || (!ex && (fn = dfnmerge(NULL, dr, pa, fi, ext)) == NULL))
	 	fatal(E_noMem);

	free(dr);
	free(pa);
	free(fi);
	free(ex);

	return fn;
}



char *mkFname(char *fname, const char *ex)
{	char *dr, *pa, *fi, *fn;

	if(!dfnsplit(fname, &dr, &pa, &fi, NULL)
	 || (fn = dfnmerge(NULL, dr, pa, fi, ex)) == NULL)
	 	fatal(E_noMem);

	free(dr);
	free(pa);
	free(fi);
	return fn;
}

char *readMsgBody(FILE *fp)
/*	read next lines, until EOF or an message header encountered */
{	char line[80];
	char *p;
	int cnt;
	int c;
#ifdef _MICROC_
	static int waiting = 0;		/* ungetc() not implemented in Micro-C */
	static char wChar;
#endif

	free(msgBody);
	msgBody = NULL;
	do {
		p = line - 1;
		cnt = 1;
		++lineNr;
		while((c = iM(waiting? (waiting = 0, wChar):) fgetc(fp)) != EOF
		 && (*++p = c) != '\n')
			if(p > line + 78) { 
				p[1] = NUL;
				addStr(msgBody, line);
				p = line - 1;
			} 
			else if(c == '>') {
				if(cnt && ++cnt == 7) {	/* message header found */
					p[-5] = NUL;	/* cut out '>>>>>>' */
					c = EOF;	/* stop reading */
					break;
				}
			}
			else {
				cnt = 0;	/* no message header */
				if(binary) {
					if(c == NUL) {	/* translate to \0 */
						featNUL = !NUL;
						*p = '\\';
#ifndef _MICROC_
						ungetc('0', fp);
#else
						waiting = wChar = '0';
#endif
					}
					else if(c == '\r') {
						/* find 'newlne' cod.
							As a newline two character strings are accepted:
								[{ \r }] \n
							and { \r }
							both are replaced by a single \n
						*/
						while((c = fgetc(fp)) == '\r');
						*p = '\n';
						if(c != '\n')
#ifndef _MICROC_
							ungetc(c, fp);
#else
							waiting = 1;	/* for c == NUL */
							wChar = c;
#endif
						break;
					}
				}
			}

		p[1] = NUL;
		addStr(msgBody, line);
	} while(c != EOF);

	/* strip last '\n' */
	if(*(p = strchr(msgBody, NUL) - 1) == '\n')
		*p = NUL;
	return msgBody;
}

char *readMsgHeader(FILE *fp)
/* treat the next line as a message header, the '>>>>>>" are already
   read */
{	static char msgHeader[40];
	char *p;
	int c;

	*(p = msgHeader) = NUL;
	while((isalnum(c = fgetc(fp)) || c == '_') && p < &msgHeader[39])
		*p++ = c;
	*p = NUL;

	while(c != '\n' && isspace(c))
		c = fgetc(fp);

	if(*msgHeader == MSG_ID_ERROR
	 || *msgHeader == MSG_ID_INTERNAL) { 	/* scan error level */
		if(isdigit(c)) {
			errNr = 0;
			do errNr = errNr * 10 + c - '0';
			while(isdigit(c = fgetc(fp)));

			while(c != '\n' && isspace(c))
				c = fgetc(fp);
		}
		else if(*msgHeader == MSG_ID_INTERNAL)
			errNr = -1;
		else
			error(E_msgFileHeadErrLvl, lineNr);
	}

	if(*msgHeader == NUL && c == EOF)
		return NULL;

	if(*msgHeader < 'A' || *msgHeader > 'Z' || msgHeader[1] != '_' 
	  || c != '\n' && c != EOF)
		error(E_msgFileHeadStruc, lineNr);

	readMsgBody(fp);	/* scan in the body and supply it in the global msgBody */
	return msgHeader;
}

void writeMsgBody(FILE *fp, unsigned char *s, int length)
/* Write the string s onto the file fp and transform all control characters
   into C-style. */
{	--s;
	if(length) do {
		if(*++s == '\t') fputs("\\t", fp);
		else if(*s == '\n') fputs("\\n\\\n", fp);
		else if(*s == '"') fputs("\\\"", fp);
		else if(*s < ' ') {
			fprintf(fp, "\\%o", *s);
			if(!*s) featNUL = !NUL;
		}
		else putc(*s, fp);
	} while(--length);
}

FILE *openDefFile(char *head, int errNr)
/* open next message definition file with nme head and error level errNr */
{	FILE *fp;

	sprintf(msgDefFileStr, msgDefFile, "", msgDefFileNr++, ".c");
	fp = Efopen(msgDefFileStr, "wt");

	message(fp, msgOpening, locHead);
	if(*head == MSG_ID_ERROR) 
		message(fp, msgErrOpening, errNr);
	message(fp, msgMsgDef, head);

	return fp;
}

void closeDefFile(FILE *fp)
/* close a message definition file */
{	fputs(msgClosing, fp);
	Efclose(fp, msgDefFileStr);
}


/*
 *	Search an (char *[]) array for a string
 *	Return:
 *		-1: not found
 *		else: index into array
 */
int searchArr(char *arr[], const char str[], unsigned size)
{	int nr;

	nr = size + 1;
	while(--nr)
		if(strcmp(*arr++, str) == 0)
			return size - nr;
	return -1;
}

int internalName(char **str)
{	int i, c;
	char *s;

	s = *str;
	if(!*s) return -1;

	while(isalpha(*s)) ++s;
	c = *s;
	*s = NUL;
	if((i = searchArr(internal, *str, keywords)) < 0)
		i = searchArr(internalE, *str, keywords);
	*s = c;

	if(i < 0) return -1;	/* error */

	while(isspace(*s)) ++s;
	*str = s;
	return i;
}

unsigned copyNumber(char *dest, char **src, char nam)
{	unsigned len;
	char *s;

	len = 0;
	s = *src;
	if(isdigit(*s)) {
		while(*s == '0') ++s;
		while(isdigit(*s)) {
			++len;
			*dest++ = *s++;
		}
		if(!len) {
			*dest++ = '0';
			++len;
		}
	}
	*dest++ = nam;
	*dest++ = NUL;
	while(isspace(*s)) ++s;
	*src = s;
	return 1 + len;
}

unsigned translateANSI(char *dest, char *src)
/* translate a ANSI sequence into its escape code representation */
{	int i, j, flag;
	char *s, ansi[32], nam;
	unsigned len;

	len = 0;
	if((s = src = strtok(src, ",")) != NULL) do {
		while(isspace(*src)) ++src;
		if((i = internalName(&s)) < 0)
			error(E_ANSIToken, src, lineNr);

#define sAnsi(a) stpcpy(ansi, a); break
#define tAnsi(a) nam = a; goto singleArg
#define cAnsi(a) nam = a; goto colourArg
#define aAnsi(a) nam = a; goto argArg
		switch(i) {
			case IN_cls: 		sAnsi("2J");
			case IN_clreol: 	sAnsi("K");
			case IN_home: 		sAnsi("H");
			case IN_save:		sAnsi("s");
			case IN_restore:	sAnsi("u");
			case IN_up:		tAnsi('A');
			case IN_down:		tAnsi('B');
			case IN_left:		tAnsi('D');
			case IN_right:		tAnsi('C');
			singleArg:
				if((i = copyNumber(ansi, &s, nam)) == 1 || i > 15)
					error(E_ANSInumericArg, src, lineNr);
				break;
			case IN_goto:
				if((i = copyNumber(ansi, &s, ';')) == 1 || i > 31
				  || (j = copyNumber(ansi + i, &s, 'H')) == i + 1 || j > 31)
					error(E_ANSIdoubleNumericArg, src, lineNr);
				break;
			default:
				flag = j = 0;	/* first colour is foreground */
				do switch(i) {
					case IN_bright:	aAnsi(1);
					case IN_blink:		aAnsi(5);
					case IN_inverse:	aAnsi(7);
					case IN_underline:	aAnsi(4);
					case IN_hidden:	aAnsi(8);
					case IN_normal:	aAnsi(0);
					case IN_black:		cAnsi(0);
					case IN_red:		cAnsi(1);
					case IN_green:		cAnsi(2);
					case IN_yellow:	cAnsi(3);
					case IN_blue:		cAnsi(4);
					case IN_magenta:	cAnsi(5);
					case IN_cyan:		cAnsi(6);
					case IN_white:		cAnsi(7);
					colourArg:
						nam |= flag? 0x40: 0x30;
						flag = 1;
					argArg:
						if(j < 29) {
							if(nam & 0xf0)
								ansi[j++] = 0x30 | (nam >> 4);
							ansi[j++] = 0x30 | (nam & 0xf);
							ansi[j++] = ';';
							break;
						}
					default:
						error(E_ANSIToken, src, lineNr);
				} while((i = internalName(&s)) >= 0);
				ansi[j - 1] = 'm';
				ansi[j] = NUL;
				break;
		}
		if(*s)
			error(E_ANSIToken, src, lineNr);
		len += 2 + strlen(ansi);
		dest = stpcpy(stpcpy(dest, "\33["), ansi);
	} while((src = s = strtok(NULL, ",")) != NULL);
	return len;
}



/*
 *	Write a message declaration
 */
void writeMsgDcl(FILE *dclFp, char *head, int group)
{	fprintf(dclFp, msgDCL, head, group, ++dclCnt);
	if(strcmp(head, "E_hlpScreen") == 0 && (dclCnt != 1 || group != 0)) 
		error(E_invalidHlpScreen);
	if(dclCnt == 0x1000)
		fatal(E_tooManyMsg);
}


/*
 *	Write the C stuff to inlcude a message
 */
void writeMsgInc(FILE *msgFp, char *p)
{	fprintf(msgFp, msgInclude, locHead, msgIncCnt, p, msgIncCnt);
	++msgIncCnt;
}


/*
 *	Write a special format id with a length number
 */
void writeSpecNum(FILE *msgFp, char id, unsigned length)
{	fprintf(msgFp, specId, id);
	fprintf(msgFp, "\\%o", length);
}

/*
 *	Write a string but chunk it into pieces of maximal 255 characters
 */
void writeChunked(FILE *msgFp, char id, char *str, int i)
{	int c;

	if(i) {			/* is there a string at all? */
		while(i > 255) {
			writeSpecNum(msgFp, id, 255);
			c = str[255];
			str[255] = NUL;
			writeMsgBody(msgFp, str, 255);
			*(str += 255) = c;
			i -= 255;
		}
		writeSpecNum(msgFp, id, i);
		writeMsgBody(msgFp, str, i);
	}
}
void writeChunkedBody(FILE *msgFp, char id, char *str)
{ writeChunked(msgFp, id, str, strlen(str)); }


/*
 *	Read a pair of words
 */
void readWrdPair(char **Xs, char **w1, char **w2)
{	char *s, *e;

	*w1 = *w2 = NULL;
	e = eow(s = skipws(*Xs));
	*w1 = strdupe(s, e);
	s = skipws(e);
	*w2 = strdupe(s, e = eol(s));
	if(!*w1 || !*w2 || !**w1)
		error(E_synSKeyword, *Xs);
	*Xs = e;
}


/*
 *	Read the global message description files and extract the internally
 *	used names into internal[] and internalE[].
 */
void readGlbFile(char *path, char *lng, FILE *dclFp)
{	char *fi, *fn, *head, *p, *h;
	FILE *fp;
	int nr;

	if(!dfnsplit(lng, NULL, NULL, &fi, NULL)
	 || (fn = dfnmerge(NULL, NULL, path, fnMsg, extMsg)) == NULL)
		error(E_noMem);

	/* Read the keyword table */
	fp = Efopen(fn, fmode);	/* global English message file */
	lineNr = 0;
	readMsgBody(fp);	/* skip header */

    while((head = readMsgHeader(fp)) != NULL)
		if(*head == MSG_ID_INTERNAL && strcmp(head, iKeywords) == 0) {
			/* message containing the English keywords. This file also
				contains the core information of the keywords:
					numerical & symbolical ID
				The message is structured like this:
				1st line: amount of keywords (numerical)
				...     : Any line starting with no digit -> comment
				Nth line: <numerical ID> <symbolical ID> <English keyword>

				The numerical ID is the index where to store the keyword and
				must be equal to the corresponding IN_???? macro.
				The symbolical ID links the keyword to the one of the
				keyword in the other languages.

				This message may appear more than once in the file, but
				the first line is ignored for all messages except the
				first one.
			*/
			
			if(!keywords) {
				if((keywords = atoi(msgBody)) <= 0)
					error(E_keywords);
				internalE = (char**)cgetmem(keywords * sizeof(char *));
				internal  = (char**)cgetmem(keywords * sizeof(char *));
			}

			head = strchr(msgBody, '\n');
			while(head && *++head) {			/* advance to next line */
				if(isdigit(*head)) {					/* keyword line */
					nr = *head - '0';
					while(isdigit(*++head))
						nr = nr * 10 + *head - '0';
					if((unsigned)nr >= keywords)
						error(E_numKeyword, nr);
					if(!isspace(*head))
						error(E_synKeyword, nr);
					readWrdPair(&head, &internal[nr], &internalE[nr]);
#ifndef NDEBUG
fprintf(stderr, "English: %u: %s -> %s\n", nr, internal[nr], internalE[nr]);
#endif
				}
				head = strchr(head, '\n');
			}
		}

	Efclose(fp, fn);
	free(fn);

	if(!keywords)
		fatal(E_noKeywords);

	for(nr = 0; nr < keywords; ++nr)
		if(!internalE[nr])
			error(E_missingInternal, nr);

	if((fn = dfnmerge(NULL, NULL, path, fi, extMsg)) == NULL)
		error(E_noMem);

	/* Read the language specific keyword table */
	fp = Efopen(fn, fmode);		/* global language specific message file */
	lineNr = 0;
	readMsgBody(fp);	/* skip header */

    while((head = readMsgHeader(fp)) != NULL) {
    	if(group != GLOBAL_GROUP)
			writeMsgDcl(dclFp, head, GLOBAL_GROUP);
    	if(*head != MSG_ID_INTERNAL)
    		continue;
		if(strcmp(head, iKeywords) == 0) {
			head = msgBody - 1;
			while(head && *++head) {			/* advance to next line */
				if(memcmp(head, symID, sizeof(symID) - 1) == 0) {
					readWrdPair(&head, &p, &h);
					if((nr = searchArr(internal, p, keywords)) < 0)
						error(E_noKeyword, p);
#ifndef NDEBUG
fprintf(stderr, "lng: %u: %s -> %s\n", nr, p, h);
#endif
					free(internal[nr]);
					internal[nr] = h;
					free(p);
				}
				head = strchr(head, '\n');
			}
		}
		else if(group != GLOBAL_GROUP && strcmp(head, iFeatures) == 0) {
			/* read the features of the global messages */
			head = msgBody - 1;
			while(head && *++head) {
				nr = 0;
				do {
					if(memcmp(featDef[nr], head, strlen(featDef[nr])) == 0) {
						features[nr] = !NUL;
#ifndef NDEBUG
fprintf(stderr, "lng: Feature: %s\n", featDef[nr]);
#endif
						break;
					}
				} while(++nr < FEATURES);
				head = strchr(head, '\n');
			}
		}
	}

	Efclose(fp, fn);
	free(fn);
	free(fi);
}

/*
 *	Dump a lower-cased defFile for use with DMake's Makefile.
 *	DMake 7-bit-lower-cases a globbed filename, but performs a case-
 *	sensitive match on that globbed name and the name within a Makefile.
 */
void lowerCaseDefFile(FILE *fp, char *prefix, int num, char *postfix)
{	char buf[sizeof(msgDefFile) + 10];

	sprintf(buf, msgDefFile, prefix, num, postfix);
	strlwr(buf);
	fputs(buf, fp);
}


/*
 *	Return the end of a special format
 *	On failure: Terminate with error message
 */
char *eoe(char *p, char id)
{
	if((p = strchr(p, id)) == NULL)
		error(E_openEncl, id, lineNr);
	return p;
}


char *opt1st = "!B";
char *opt2nd = "CGLM";
main(int argc, char **argv)
{	char *head, *tail, *p;
	FILE *fp, *dclFp, *msgFp;
	int c;

    char  *path				/* path to the global message files */
    	 ,*lng				/* language to be compiled */
    	 ,*cRspFn			/* compiler response file */
    	 ,*lRspFn			/* librarian response file */
    	 ,*mRspFn			/* Makefile dependency file */
    	 ,*featFn			/* used feature logfilename */
    	 ,*dclFn			/* message declaration filename */
    	 ;

    unsigned i;

    msgInit();

	cRspFn = lRspFn = mRspFn = featFn = dclFn = NULL;

	while((int)(i = getoptG(argc, argv)) != EOF) switch(i) {
	default:	hlpScreen();
	case '!':				/* use msg!#### instead of msg_#### */
		msgDefFile[padDefFile] = '!';
		group = GLOBAL_GROUP;
		break;
	case 'B':				/* use binary mode */
		binary = !binary;
		break;
	case 'C':				/* compiler response file */
		cRspFn = optarg;
		break;
	case 'L':				/* librarian response file */
		lRspFn = optarg;
		break;
	case 'M':				/* Makefile dependencies */
		mRspFn = optarg;
		break;
	case 'G':				/* message group */
		if((group = atoi(optarg)) < 0 || group > 15)
			error(E_rangeOfGroup);
		break;
	}

	if(binary)								/* read in binary mode */
		fmode[1] = 'b';

	switch(argc - optind) {
	default:	hlpScreen();
	case 5:		featFn = argv[optind + 4];
	case 4:		dclFn = argv[optind + 3];
	case 3:		locHead = argv[optind + 2];
				lng = argv[optind + 1];
				path = argv[optind + 0];
				break;
	}

	if(!dclFn)
		dclFn = mkFname(lng, extDcl);
	dclFp = Efopen1(dclFn);

	/* read internally used keywords and the features and declarations
		of the global message file */
	informative(M_readInternal);
	readGlbFile(path, lng, dclFp);


	/* read message file & create the message definition files */
	informative(M_createDefFiles);
	lng = attachExt(lng, extMsg);
	fp = Efopen(lng, fmode);

	dclCnt = lineNr = 0;
	readMsgBody(fp);	/* skip header */

	while((head = readMsgHeader(fp)) != NULL) {
		writeMsgDcl(dclFp, head, group);
		msgFp = openDefFile(head, errNr);
		p = (tail = msgBody) - 1;
		while((p = strchr(p + 1, '%')) != NULL) {	/* specil format spec */
			if(p != tail) {			/* emit previous text */
				writeMsgBody(msgFp, tail, p - tail);
				tail = p;
			}

			/* tail points to the opening percent sign '%', p to the
				working place behind it */

			if(*++p == '%') {	/* skip %% */
				continue;
			}
			if(*p == '-') {
				if(p[1] == '%') {		/* %-% found : ANSI enclosure */
					featAnsi = !NUL;
					*(p = eoe(p + 2, '%')) = NUL;
					writeChunked(msgFp, 'A', tail
					 , translateANSI(tail, tail + 3));
					tail = p + 1;
					continue;
				}
				else if(p[1] == '-') {
					if(p[2] == '%') {
						switch(*(p += 3)) {
						case 'A':
							featAnsi = !NUL;
							goto copyChunk;
						case 'W':
							featRaw = !NUL;
						copyChunk:
							c = *++p;
							*(p = eoe(tail = p, c)) = NUL;
							writeChunkedBody(msgFp, 'A', tail);
							tail = p + 1;
							break;
						case 'r':				/* %--%r?###% */
							featRep = !NUL;
							tail = eoe(p + 2, '%') + 1;
							if(!(unsigned)(i = atoi(p + 2)))
								error(E_numberZero);
							while((unsigned)i > 255) {
								writeSpecNum(msgFp, 'r', 255);
								putc(p[1], msgFp);
								i -= 255;
							}
							if(i) {
								writeSpecNum(msgFp, 'r', i);
								putc(p[1], msgFp);
							}
							break;
						case 'R':				/* %--%R###% */
							featRepArg = !NUL;
							tail = eoe(p + 1, '%') + 1;
							if((unsigned)(i = atoi(p + 1)) > 255 || !i)
								error(E_invalidSpec, 'R');
							writeSpecNum(msgFp, 'R', i);
							break;
						case 'm':				/* %--%m****% */
							featMInc = !NUL;
							tail = eoe(++p, '%');
							*tail++ = NUL;
							fprintf(msgFp, specId, 'm');
							writeMsgInc(msgFp, p);
							break;
						case 'M':				/* %--%M */
							featMIncArg = !NUL;
							fprintf(msgFp, specId, 'M');
							tail = p + 1;
							break;
						}
					}
					continue;
				}
				++p;
			}
			while(isdigit(*p)) ++p;
			if(*p == '%') {
				--p;
				featReorder = !NUL;
			}
		}
		writeMsgBody(msgFp, tail, strlen(tail));
		closeDefFile(msgFp);
	}
	Efclose(fp, lng);
	Efclose(dclFp, dclFn);

	/* create the compiler response file */
	if(cRspFn) {	/* response file */
		informative(M_createCompRsp);
		fp = Efopen1(cRspFn);
		for(i = 0; i < msgDefFileNr; ++i)
			fprintf(fp, msgDefFile, "", i, ".c\n");
		Efclose(fp, cRspFn);
	}
	else warning(W_compRsp);

	/* create the librarian response file */
	if(lRspFn) {	/* response file */
		informative(M_createLibRsp);
		fp = Efopen1(lRspFn);
		head = "+ ";
		for(i = 0; i < msgDefFileNr; ++i)
			fprintf(fp, msgDefFile, head, i, ""), head = ".obj &\n+ ";
		if(*head == '.') fputs(".obj\n", fp);
		Efclose(fp, lRspFn);
	}
	else warning(W_libRsp);

	/* create the Makefile dependency file */
	if(mRspFn) {	/* response file */
		informative(M_createMakedep);
		strlwr(locHead);		/* for DMake's case-sensitive math against
									lower-cased globbed filenames. */
		fp = Efopen1(mRspFn);
		fputs("all : ", fp);	/* create the default target */
		for(i = 0; i < msgDefFileNr; ++i)
			lowerCaseDefFile(fp, " \\\n\t", i, ".obj");
		fputs("\n\n\n", fp);
		for(i = 0; i < msgDefFileNr; ++i) {
			lowerCaseDefFile(fp, "", i, ".obj");
			lowerCaseDefFile(fp, " : ", i, ".c ");
			fputs(locHead, fp);
			putc('\n', fp);
			putc('\n', fp);
		}
		Efclose(fp, mRspFn);
	}
	else warning(W_makRsp);

	/* create the feature file */
	if(!featFn)
		featFn = mkFname(lng, extFtr);
	informative(M_createFeature, featFn);
	fp = Efopen1(featFn);
	c = 0;
	do {
		if(features[c]) {
			fputs(featDef[c], fp);
			putc('\n', fp);
		}
	} while(++c < FEATURES);
	Efclose(fp, featFn);

#ifndef _MICROC_
	return 0;	/* standard for Micro-C */
#endif
}
